import asyncio
import logging
from json import JSONDecodeError
from typing import Dict, Text, Any, Optional

import aiohttp
from aiohttp import web, WSMsgType, WSMessage
from aiohttp.web_request import Request
from aiohttp.web_ws import WebSocketResponse

logger = logging.getLogger(__name__)

CONNECTED_WEB_SOCKETS = {}  # type: Dict[Text, WebSocketResponse]
DEFAULT_WEBSOCKET_URL = "ws://localhost:5003/ws"


class NotificationService:
    def __init__(self, websocket_url: Text = DEFAULT_WEBSOCKET_URL):
        self._websocket_url = websocket_url

        # Lazy initialization of the connections
        self._session = None  # type: Optional[aiohttp.ClientSession]
        self._websocket = None  # type: Optional[aiohttp.ClientWebSocketResponse]

    def __del__(self) -> None:
        # Close everything when the object get's destroyed.
        asyncio.ensure_future(self._close_websocket())

    async def _close_websocket(self) -> None:
        if self._websocket:
            await self._websocket.close()

        if self._session:
            await self._session.close()

    async def _connect_websocket(self) -> None:
        if not self._session:
            self._session = aiohttp.ClientSession()

        if not self._websocket:
            self._websocket = await self._session.ws_connect(self._websocket_url)

    async def _send_notification(self, message: Dict[Text, Any]) -> None:
        """Send a message to the websocket server.

        Args:
            message: The message.
        """

        await self._connect_websocket()
        await self._websocket.send_json(message)

    async def send_notification_to(
        self,
        recipient_id: Optional[Text],
        topic: Text,
        status: Text,
        details: Optional[Dict[Text, Any]] = None,
    ) -> None:
        """Send a notification to a Rasa X user if the user has a connected web socket.

        Args:
            recipient_id: the id of the Rasa X user which should get the notification.
            topic: the category of the message
            status: status of the operation the notification describes.
            details: optional additional information of the notification.
        """

        if not recipient_id:
            # Handling the `None` case makes it way easier for callers of the function
            # which may or may have not the recipient
            return

        message = {
            "topic": topic,
            "status": status,
            "details": details,
            "recipient_id": recipient_id,
        }
        await self._send_notification(message)


def run_websocket_server() -> None:
    """Runs a websocket server."""

    app = web.Application()
    app.add_routes([web.get("/ws", _websocket_handler)])
    logger.debug("Starting web socket server on '{}'.".format(DEFAULT_WEBSOCKET_URL))

    web.run_app(app, port=5003)


async def _websocket_handler(request: Request) -> WebSocketResponse:
    """Handles new connections to the websocket server."""

    websocket = web.WebSocketResponse()

    await websocket.prepare(request)

    async for message in websocket:  # type: WSMessage
        if message.type == WSMsgType.TEXT:
            parsed_message = _get_json(message)

            if _is_from_rasa_x_server(request):
                recipient_id = parsed_message.get("recipient_id")

                if recipient_id:
                    await _forward_message_to(recipient_id, parsed_message)
                else:
                    await _forward_broadcast_to_connected_websockets(parsed_message)

            # Don't differ between localhost and other hosts since the UI could also
            # run on localhost
            _handle_user_messages(message, websocket)

    _remove_websocket_connection(websocket)
    return websocket


def _get_json(message: WSMessage) -> Dict:
    try:
        return message.json() or {}
    except JSONDecodeError:
        return {}


def _is_from_rasa_x_server(request: Request) -> bool:
    return request.host.startswith("localhost")


async def _forward_message_to(recipient_id: Text, message: Dict[Text, Any]) -> None:
    websocket_of_recipient = CONNECTED_WEB_SOCKETS.get(recipient_id)
    if websocket_of_recipient is not None:
        await websocket_of_recipient.send_json(message)
        logger.debug("Sent notification to recipient '{}'.".format(recipient_id))
    else:
        logger.debug(
            "Web socket connection for recipient '{}' not found.".format(recipient_id)
        )


async def _forward_broadcast_to_connected_websockets(message: Dict[Text, Any]) -> None:
    logger.debug("Broadcasting this message: {}".format(message))
    send_message_coroutines = [
        w.send_json(message) for w in CONNECTED_WEB_SOCKETS.values()
    ]
    await asyncio.gather(*send_message_coroutines)


def _handle_user_messages(message: WSMessage, websocket: WebSocketResponse):
    content = _get_json(message)
    # Users have to sent a message with their username to get added to the connections.
    username = content.get("username")
    if username:
        _add_websocket_connection(username, websocket)


def _add_websocket_connection(sender_id: Text, websocket: WebSocketResponse) -> None:
    CONNECTED_WEB_SOCKETS[sender_id] = websocket
    logger.debug("New web socket connection with '{}'.".format(sender_id))


def _remove_websocket_connection(websocket_to_delete: WebSocketResponse) -> None:
    matching_keys = [
        sender_id
        for sender_id, socket in CONNECTED_WEB_SOCKETS.items()
        if socket == websocket_to_delete
    ]
    for sender_id in matching_keys:
        del CONNECTED_WEB_SOCKETS[sender_id]
        logger.debug(
            "Removed websocket for '{}' from connected web sockets.".format(sender_id)
        )
